(define-module (kjartan home services qemu)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-28)
  #:use-module (srfi srfi-171)
  #:use-module (ice-9 match)
  #:use-module (guix gexp)
  #:use-module (guix records)
  #:use-module (guix packages)
  #:use-module (gnu services)
  #:use-module (gnu services configuration)
  #:use-module (gnu home services shepherd)
  #:use-module (gnu home services)
  #:use-module (gnu packages virtualization)
  #:export (qemu-vm-configuration
            qemu-spice-configuration
            qemu-shared-directory
            home-qemu-configuration
            home-qemu-service-type))

(define (positive-integer? val)
  (and (exact-integer? val) (not (negative? val))))

(define (port-number? val)
  (and (exact-integer? val) (not (negative? val))))

(define (port-forwarding-rule? val)
  (and (pair? val) (port-number? (car val)) (port-number? (cdr val))))

(define (list-of-port-forwarding-rules? val)
  (every port-forwarding-rule? val))

(define (security-model? val)
  (memq val '(passthrough mapped-xattr mapped-file none)))

(define-configuration/no-serialization qemu-shared-directory
  (path string "The path to the directory")
  (tag string "The mount tag of the directory")
  (security-model security-model "The security model of the directory, one of passthrough, mapped-xattr, mapped-file, none"))

(define (list-of-shared-directories? val)
  (every qemu-shared-directory? val))

(define (shared-directory-arguments directory)
  `("-virtfs"
    ,(format "local,path=~a,mount_tag=~a,security_model=~a"
             (qemu-shared-directory-path directory)
             (qemu-shared-directory-tag directory)
             (qemu-shared-directory-security-model directory))))

(define-configuration/no-serialization qemu-spice-configuration
  (port port-number "The port for the SPICE connection"))

(define (graphics-configuration? val)
  (match val
    ('none #t)
    ('nographic #t)
    ((? qemu-spice-configuration? val) #t)
    (_ #f)))

(define (graphics-arguments config)
  (match config
    ('none '("-display" "none"))
    ('nographic '("-nographic"))
    ((? qemu-spice-configuration? config)
     `("-display" "-spice" ,(format "port=~a" (qemu-spice-configuration-port config))))))

(graphics-configuration? (qemu-spice-configuration
                          (port 3030)))

(graphics-arguments (qemu-spice-configuration
                          (port 3030)))

(define-configuration/no-serialization qemu-vm-configuration
  (name symbol "The name of the VM")
  (documentation string "Documentation for the VM")
  (memory string "The amount of RAM the VM should have.  Accepts suffixes like M or G to
specify Mega- or Gigabytes.")
  (cpus (positive-integer 1) "The number of CPU cores the VM should have.")
  (disk-image string "The disk image the VM should use.")
  (kvm? (boolean #f) "Whether to enable KVM.")
  (graphics graphics-configuration "The graphics configuration for the VM.")
  (forward-ports (list-of-port-forwarding-rules '()) "The list of ports to forward to the VM.  Each entry is a pair (HOST . GUEST)")
  (shared-directories (list-of-shared-directories '()) "The list of directories to share with the VM.")
  (auto-start? (boolean #f) "Whether the VM should be started automatically.")
  (extra-args (list-of-strings '()) "Other arguments to pass directly to QEMU."))

(define (qemu-vm-shepherd-service-factory qemu)
  (let ((bin (file-append qemu "/bin/qemu-system-x86_64")))
    (lambda (config)
      (shepherd-service
       (documentation (qemu-vm-configuration-documentation config))
       (provision (list (qemu-vm-configuration-name config)))
       (auto-start? (qemu-vm-configuration-auto-start? config))
       (start
        #~(make-forkexec-constructor
           '#$(append
               `(,bin
                 "-hda" ,(qemu-vm-configuration-disk-image config)
                 "-m" ,(qemu-vm-configuration-memory config)
                 "-smp" ,(number->string (qemu-vm-configuration-cpus config)))
               (if (qemu-vm-configuration-kvm? config)
                   '("--enable-kvm")
                   '())
               (graphics-arguments (qemu-vm-configuration-graphics config))
               (if (qemu-vm-configuration-forward-ports config)
                   (list-transduce
                    tflatten rcons
                    (map (lambda (rule)
                           (list "-nic" (format "user,hostfwd=tcp::~a-:~a" (car rule) (cdr rule))))
                         (qemu-vm-configuration-forward-ports config)))
                   '())
               (if (qemu-vm-configuration-shared-directories config)
                   (list-transduce
                    tflatten rcons
                    (map shared-directory-arguments
                         (qemu-vm-configuration-shared-directories config))))
               (qemu-vm-configuration-extra-args config))))
       (stop #~(make-kill-destructor))))))

(define (list-of-qemu-vm-configurations? lst)
  (every qemu-vm-configuration? lst))

(define-configuration/no-serialization home-qemu-configuration
  (qemu
   (package qemu)
   "The qemu package to use")
  (virtual-machines
   (list-of-qemu-vm-configurations '())
   "A list of VM configurations."))

(define (qemu-shepherd-service config)
  (map (qemu-vm-shepherd-service-factory (home-qemu-configuration-qemu config))
       (home-qemu-configuration-virtual-machines config)))

(define home-qemu-service-type
  (service-type
   (name 'qemu)
   (extensions
    (list
     (service-extension home-shepherd-service-type
                        qemu-shepherd-service)))
   (description "A Home service for running QEMU VMs.")))
